/*
	PROforma, PROGS FOnt & Raster MAnager

	Routine to convert a character description into a sequence of commands
	to draw that character (including the width).

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 July 1994, based on PFShowChar in v1.02 of PROforma
	09.10.1995 changed naming scheme
	12.10.1995 adjusted because caching mechanism changed
	16.10.1995 removed some unnecessary tests on currfont->glyphs
	17.10.1995 support for builtin font with switching for non existing
			   characters + charpath works again for clients
	18.10.1995 PFCharPathEl now returns absolute coordinates
			   the width is always returned at real font size
	22.10.1996 CHARLOG defines for character logging added
			   this should allow me to improve font quality
	24.10.1996 logging now under some kind of mutex
	01.11.1996 curve now handled differently cfr. better hinting. The
			   controlpoints are not realtive with the endpoints, but are
			   converted sompletely seperate
	29.11.1996 calculating of outline for rendering characters integrated here
	26.12.1996 fixed width fonts now remain fixed width when using PFP routs
	20.02.1997 bug in PFCharPathInit fixed currfont was not filled in
	02.08.1997 ateempt to solve the 'm' problem
*/

#include <job.h>
#include <array.h>
#include <atomic.h>

#include "gstate.h"
#include "system.h"
#include "path.h"
#include "hint.h"
#include "cache.h"

#ifdef CHARLOG
Channel log=CHANNEL_NONE;
static char *logging=NULL;
static int logcount=0;
#endif CHARLOG

extern font *ResFonts;				/* pointer to builting font */

static char notdef=fc_endchar;

/* table to convert StandardEncoding to Unicode */
static const short StandardEncoding[256]={
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 32, 33, 34, 35, 36, 37, 38, 8217, 40, 41,
		42, 43, 44, 45, 46, 47, 48, 49, 50, 51,
		52, 53, 54, 55, 56, 57, 58, 59, 60, 61, 62, 63,
		64, 65, 66, 67, 68, 69, 70, 71, 72, 73, 74, 75,
		76, 77, 78, 79, 80, 81, 82, 83, 84, 85, 86, 87,
		88, 89, 90, 91, 92, 93, 94, 95, 8216, 97, 98, 99,
		100, 101, 102, 103, 104, 105, 106, 107, 108, 109,
		110, 111, 112, 113, 114, 115, 116, 117, 118, 119,
		120, 121, 122, 123, 124, 125, 126, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 161, 162, 163,
		8725, 165, 402, 167, 164, 39, 8220, 171, 8249,
		8250, (short)0xf001, (short)0xf002, 1, 8211, 8224, 8225, 8729, 1,
		182, 8226, 8218, 8222, 8221, 187, 8230, 8240, 1, 191,
		1, 96, 180, 710, 732, 713, 728, 729, 168, 1, 730,
		184, 1, 733, 731, 711, 8212, 1, 1, 1, 1, 1, 1, 1, 1,
		1, 1, 1, 1, 1, 1, 1, 1, 198, 1, 170, 1, 1, 1, 1, 321,
		216, 338, 186, 1, 1, 1, 1, 1, 230, 1, 1, 1,
		305, 1, 1, 322, 248, 339, 223, 1, 1, 1, 1
};

/* compare routine for binary search */
static int compare(const void *a, const void *b)
{
	return ((long)(((ref *)a)->code)&0xffff) -
		   ((long)(((ref *)b)->code)&0xffff);
}

/* routine to return the start of the given character */
static char *charstart(Gstate gstate, short unicode)
{
	/* search the given character code with a binary search */
	/* or return the first character in the list (.notdef) if not found */
	Error err;
	ref *rf, r;
	CharPath *chp=&gstate->chp;
	r.code=unicode;

	if (!chp->currfont)
	{
		chp->hints=&gstate->font->hints;
		chp->currfont=gstate->currfont;
	}

	err=ARRAYSearch(&r, chp->currfont->glyphs->list,
						chp->currfont->glyphs->count,
						sizeof(ref), compare, &rf);
	if (err)
	{
		/* not found -> search character in builtin font */
		chp->hints=&gstate->font->bltin;
		chp->currfont=ResFonts;

		err=ARRAYSearch(&r, ResFonts->glyphs->list,
							ResFonts->glyphs->count,
							sizeof(ref), compare, &rf);
	}

	/* handle plausible errors in font file */
	if (err || rf->code==0 || rf->base==0) return &notdef;

	return ((int)chp->currfont->glyphs) + rf->base;
}

/* routine to return the start of the subroutine character */
static char *subrstart(CharPath *chp, short code)
{
	/* search the given subroutine with a binary search */
	ref *rf, r;
	r.code=code;

	ARRAYSearch(&r, chp->currfont->subrs->list,
					chp->currfont->subrs->count,
					sizeof(ref), compare, &rf);
	return ((int)chp->currfont->subrs) + rf->base;
}

/* routine to get the next command */
#define getcommand command=*chp->charpos++

Error PFCharPathInit(Gstate gstate, int code)
{
	CharPath *chp=&gstate->chp; 	  /* data for this character */

	/* make sure everything works in case of an error */
	MEMClear(sizeof(CharPath),chp);

	/* see if we can use the hinting mechanism */
	chp->hints=&chp->clienthints;
	chp->currfont=gstate->currfont;
	MEMClear(sizeof(hintdata),chp->hints);
	hintMatrix(gstate,
			   (CTM *)chp->currfont->info->FontMatrix, &chp->hints->matrix);

	/* find base of character */
	chp->charpos=charstart(gstate, code);
	HintClear(chp->hints);

	/* initialise charpath */
	chp->currstack=chp->stack;			/* current pos in font stack */
	chp->currretstack=chp->retstack;	/* current pos in subrout stack */
	chp->seac=chp->contchar=FALSE;		/* no seac yet */
	chp->rendering=FALSE;

	return ERR_OK;
}

#ifdef CHARLOG
static void clearlogtest()
{
	logging=NULL;
}
static int logtest()
{
	if (logging)
		return TRUE;
	MEMAllocate(1,&logging);
	MEMWhenZapped(logging,clearlogtest,NULL);
	return FALSE;
}
#endif CHARLOG

Error CalcChar(Gstate gstate, short code)
{
	CharPath *chp=&gstate->chp; 	  /* data for this character */
	short command;

	/* make sure everything works in case of an error */
	MEMClear(sizeof(CharPath),chp);

	/* find base of character */
	chp->charpos=charstart(gstate, code);
	HintClear(chp->hints);

	/* initialise charpath */
	chp->currstack=chp->stack;			/* current pos in font stack */
	chp->currretstack=chp->retstack;	/* current pos in subrout stack */
	chp->seac=chp->contchar=FALSE;		/* no seac yet */

	gstate->CharWidth.x=gstate->CharWidth.y=0;
	gstate->boxxmin=gstate->boxymin=ptmax;
	gstate->boxxmax=gstate->boxymax=ptmin;
	chp->rendering=TRUE;
	chp->code=code;
	chp->gstate=gstate;
	chp->pathorg.x=chp->pathorg.y=chp->currpos.x=chp->currpos.y=0;

#ifdef CHARLOG
{
	char line[80];
	/* some quick and dirty mutex */
	while (ATOMICCall(logtest,0)) JOBSuspend(MYSELF,1,NULL);
	if (log==CHANNEL_NONE) IOOpen("win1_fontlog",OPEN_OVER,&log);
	STRFormatPut(line,"----- start of char -- %d ",logcount++);
	IOPutString(log,line,NULL);
}
#endif CHARLOG

	return PFCharPathEl(gstate,&command,NULL,NULL,NULL,NULL,NULL,NULL);
}

/* here are the actual routines which set the return values */
static int do_move(CharPath *chp, pt **data, int rx, int ry)
{
	coor c;
	Gstate gstate=chp->gstate;
#ifdef CHARLOG
	char line[80];
	pt **d=data;
	STRFormatPut(line,"MOVE %d %d",rx,ry);
	IOPutString(log,line,NULL);
#endif CHARLOG

	c.x=chp->charposx+=rx; c.y=chp->charposy+=ry;
	font2dev(chp->hints,&c);
	if (chp->rendering)
	{
		addflatline(gstate,chp->currpos.x,chp->currpos.y,chp->pathorg.x,chp->pathorg.y);
		addflatline2(gstate,chp->currpos.y,chp->currpos.x,chp->pathorg.y,chp->pathorg.x);
		chp->pathorg.x=chp->currpos.x=c.x;
		chp->pathorg.y=chp->currpos.y=c.y;
		newmin(gstate->boxxmin,c.x); newmax(gstate->boxxmax,c.x);
		newmin(gstate->boxymin,c.y); newmax(gstate->boxymax,c.y);
	}
	else
	{
		*(short *)(*data++)=PF_CHARPATH_MOVETO;
		**data++=c.x;
		**data++=c.y;
	}
#ifdef CHARLOG
	STRFormatPut(line,"%d %f %f",PF_CHARPATH_MOVETO,pt2double(c.x),pt2double(c.y));
	IOPutString(log,line,NULL);
#endif CHARLOG
	chp->currstack=chp->stack;
	return !chp->rendering;
}

static int do_line(CharPath *chp, pt **data, int rx, int ry)
{
	coor c;
	Gstate gstate=chp->gstate;
#ifdef CHARLOG
	char line[80];
	pt **d=data;
	STRFormatPut(line,"LINE %d %d",rx,ry);
	IOPutString(log,line,NULL);
#endif CHARLOG

	c.x=chp->charposx+=rx; c.y=chp->charposy+=ry;
	font2dev(chp->hints,&c);
	if (chp->rendering)
	{
		addflatline(gstate,chp->currpos.x,chp->currpos.y,c.x,c.y);
		addflatline2(gstate,chp->currpos.y,chp->currpos.x,c.y,c.x);
		chp->currpos.x=c.x;
		chp->currpos.y=c.y;
		chp->draw=TRUE;
		newmin(gstate->boxxmin,c.x); newmax(gstate->boxxmax,c.x);
		newmin(gstate->boxymin,c.y); newmax(gstate->boxymax,c.y);
	}
	else
	{
		*(short *)(*data++)=PF_CHARPATH_LINETO;
		**data++=c.x;
		**data++=c.y;
	}
#ifdef CHARLOG
	STRFormatPut(line,"%d %f %f",PF_CHARPATH_LINETO,pt2double(c.x),pt2double(c.y));
	IOPutString(log,line,NULL);
#endif CHARLOG
	chp->currstack=chp->stack;
	return !chp->rendering;
}

#define do_aline(chp,d,rx,ry) { chp->charposx=chp->charposy=0; do_line(chp,d,rx,ry); }

static int do_curve(CharPath *chp, pt **data, int rx1, int ry1,
					int rx2, int ry2, int rx, int ry)
{
	coor c,c1,c2;
	Gstate gstate=chp->gstate;
#ifdef CHARLOG
	char line[80];
	pt **d=data;
	STRFormatPut(line,"CURVE %d %d %d %d %d %d",rx1,ry1,rx2,ry2,rx,ry);
	IOPutString(log,line,NULL);
#endif CHARLOG

	c1.x=chp->charposx+rx1;
	c1.y=chp->charposy+ry1;
	c2.x=c1.x+rx2; c2.y=c1.y+ry2;
	c.x=chp->charposx=c2.x+rx;
	c.y=chp->charposy=c2.y+ry;
	font2dev(chp->hints,&c1);
	font2dev(chp->hints,&c2);
	font2dev(chp->hints,&c);
	if (chp->rendering)
	{
		addflatcurve(gstate,chp->currpos.x,chp->currpos.y,
					 c1.x,c1.y,c2.x,c2.y,c.x,c.y,
					 pt_one/16, addflatline, (Bitmap)gstate);
		addflatcurve(gstate,chp->currpos.y,chp->currpos.x,
					 c1.y,c1.x,c2.y,c2.x,c.y,c.x,
					 pt_one/16, addflatline2, (Bitmap)gstate);
		chp->currpos.x=c.x;
		chp->currpos.y=c.y;
		newmin(gstate->boxxmin,c1.x); newmax(gstate->boxxmax,c1.x);
		newmin(gstate->boxymin,c1.y); newmax(gstate->boxymax,c1.y);
		newmin(gstate->boxxmin,c2.x); newmax(gstate->boxxmax,c2.x);
		newmin(gstate->boxymin,c2.y); newmax(gstate->boxymax,c2.y);
		newmin(gstate->boxxmin, c.x); newmax(gstate->boxxmax, c.x);
		newmin(gstate->boxymin, c.y); newmax(gstate->boxymax, c.y);
		chp->draw=TRUE;
	}
	else
	{
		*(short *)(*data++)=PF_CHARPATH_CURVETO;
		**data++=c1.x;
		**data++=c1.y;
		**data++=c2.x;
		**data++=c2.y;
		**data++=c.x;
		**data++=c.y;
	}
#ifdef CHARLOG
	STRFormatPut(line,"%d %f %f %f %f %f %f",PF_CHARPATH_CURVETO,
		pt2double(c1.x),pt2double(c1.y),
		pt2double(c2.x),pt2double(c2.y),
		pt2double( c.x),pt2double( c.y));
	IOPutString(log,line,NULL);
#endif CHARLOG
	chp->currstack=chp->stack;
	return !chp->rendering;
}

static int do_endchar(Gstate gstate, CharPath *chp)
{
	if (chp->rendering)
	{
		CachedFont *font=gstate->font;
		short code=chp->code;

		addflatline(gstate,chp->currpos.x,chp->currpos.y,chp->pathorg.x,chp->pathorg.y);
		addflatline2(gstate,chp->currpos.y,chp->currpos.x,chp->pathorg.y,chp->pathorg.x);
		if (chp->draw)
		{
			gstate->boxxmin=pt2short(gstate->boxxmin-pt_half);
			gstate->boxymin=pt2short(gstate->boxymin-pt_half);
			gstate->boxxmax=pt2short(gstate->boxxmax+pt_half);
			gstate->boxymax=pt2short(gstate->boxymax+pt_half);
		}
		else			/* handle characters which contain no drawing commands */
		{
			gstate->boxxmin=
			gstate->boxymin=0;
			gstate->boxxmax=
			gstate->boxymax=1;
			if (code==' ')
				gstate->boxxmax=pt2rshort(
								fixmul(gstate->CharWidth.x,gstate->todev.x));
		}
		if (code=='m')
			font->mwidth=gstate->CharWidth.x;
		else
		{
			if (font->mwidth==gstate->CharWidth.x)
				gstate->boxxmax=font->pwidth.array['m'-PF_WIDTHARRAY_FIRST]-1;
		}
		if (code<=PF_WIDTHARRAY_LAST && code>=PF_WIDTHARRAY_FIRST)
			font->pwidth.array[code-PF_WIDTHARRAY_FIRST]=gstate->boxxmax+1;
	}
	return TRUE;
}

static int do_width(Gstate gstate, pt **data, int wx, int wy)
{
	coor c;
	CharPath *chp=&gstate->chp;
#ifdef CHARLOG
	char line[80];
	pt **d=data;
	STRFormatPut(line,"WIDTH %d %d",wx,wy);
	IOPutString(log,line,NULL);
#endif CHARLOG

	c.x=wx; c.y=wy;
	widthcoor(gstate,&c);
	if (chp->rendering)
	{
		gstate->CharWidth.x=c.x;
		gstate->CharWidth.y=c.y;
	}
	else
	{
		*(short *)(*data++)=PF_CHARPATH_WIDTH;
		**data++=c.x;
		**data++=c.y;
	}
#ifdef CHARLOG
	STRFormatPut(line,"%d %f %f",PF_CHARPATH_WIDTH,pt2double(c.x),pt2double(c.y));
	IOPutString(log,line,NULL);
#endif CHARLOG
	chp->currstack=chp->stack;
	return !chp->rendering;
}


/* internal routine to get the details of a CharPath, return absolute coors */
Error PFCharPathEl(Gstate gstate, short *comm, pt *x1, pt *y1, pt *x2, pt *y2, pt *x3, pt *y3)
{
	CharPath *chp=&gstate->chp; 	  /* data for this character */
	unsigned char command;
	char endchar=FALSE;
	char flex=FALSE;
	long value;
#ifdef CHARLOG
	char line[100];
#endif CHARLOG

	*comm=PF_CHARPATH_END;
	if (chp->charpos==NULL) return ERR_OK;

	/* process character */
	while (!endchar)
	{
		getcommand;
		if (command>=minval)
		{
			if (command<=maxval)
			{
				*chp->currstack++=command-139;
			} else if (command<=maxposword)
			{
				*chp->currstack++=(command-minposword)*256 +
								  ((unsigned char)*chp->charpos++) + 108;
			} else if (command<=maxnegword)
			{
				*chp->currstack++=-(command-minnegword)*256 -
								  ((unsigned char)*chp->charpos++) - 108;
			}
			else
			{
				MEMCopy(4,chp->charpos,&value); /* !!!!!!!!!!!!! tricky !!!!!!!!! */
				chp->charpos+=4;				/* sequence needs big endian long */
				*chp->currstack++=value;
			}
		} else switch(command)
		{
		case fc_hstem:
#ifdef CHARLOG
	STRFormatPut(line,"HSTEM %d %d",chp->stack[0]+chp->sby,chp->stack[1]);
	IOPutString(log,line,NULL);
#endif CHARLOG
			HintY(chp->hints,chp->stack[0]+chp->sby,chp->stack[1]);
			chp->currstack=chp->stack;			  /* clear char operand stack */
			break;

		case fc_vstem:
#ifdef CHARLOG
	STRFormatPut(line,"VSTEM %d %d",chp->stack[0]+chp->sbx,chp->stack[1]);
	IOPutString(log,line,NULL);
#endif CHARLOG
			HintX(chp->hints,chp->stack[0]+chp->sbx,chp->stack[1]);
			chp->currstack=chp->stack;			  /* clear char operand stack */
			break;

		case fc_closepath:
			/* closepath; doesn't need any processing !! */
			chp->currstack=chp->stack;			  /* clear char operand stack */
			break;

		case fc_rlineto:
			endchar=do_line(chp,(pt **)&comm,chp->stack[0],chp->stack[1]);
			break;

		case fc_hlineto:
			endchar=do_line(chp,(pt **)&comm,chp->stack[0],0);
			break;

		case fc_vlineto:
			endchar=do_line(chp,(pt **)&comm,0,chp->stack[0]);
			break;

		case fc_callsubr:
			*chp->currretstack++=(long)chp->charpos;
			chp->charpos=subrstart(chp, *--chp->currstack);
			break;

		case fc_return:
			chp->charpos=(char *)(*--chp->currretstack);
			break;

		case fc_escape:
			getcommand;
			switch(command)
			{
			case fec_dotsection:
				chp->currstack=chp->stack;			  /* clear char operand stack */
				break;

			case fec_vstem3:
#ifdef CHARLOG
	STRFormatPut(line,"vSTEM3 %d %d %d %d %d %d",
				chp->stack[0]+chp->sbx,chp->stack[1],
				chp->stack[2]+chp->sbx,chp->stack[3],
				chp->stack[4]+chp->sbx,chp->stack[5]);
	IOPutString(log,line,NULL);
#endif CHARLOG
				HintX(chp->hints,chp->stack[0]+chp->sbx,chp->stack[1]);
				HintX(chp->hints,chp->stack[2]+chp->sbx,chp->stack[3]);
				HintX(chp->hints,chp->stack[4]+chp->sbx,chp->stack[5]);
				chp->currstack=chp->stack;			  /* clear char operand stack */
				break;

			case fec_hstem3:
#ifdef CHARLOG
	STRFormatPut(line,"hSTEM3 %d %d %d %d %d %d",
				chp->stack[0]+chp->sby,chp->stack[1],
				chp->stack[2]+chp->sby,chp->stack[3],
				chp->stack[4]+chp->sby,chp->stack[5]);
	IOPutString(log,line,NULL);
#endif CHARLOG
				HintY(chp->hints,chp->stack[0]+chp->sby,chp->stack[1]);
				HintY(chp->hints,chp->stack[2]+chp->sby,chp->stack[3]);
				HintY(chp->hints,chp->stack[4]+chp->sby,chp->stack[5]);
				chp->currstack=chp->stack;			  /* clear char operand stack */
				break;

			case fec_seac:
#ifdef TESTFONT
printf("seac: %d %d %d %d %d seac\n",chp->stack[0],chp->stack[1],chp->stack[2],chp->stack[3],
	chp->stack[4],chp->stack[5]);
#endif
				chp->charpos=charstart(gstate, StandardEncoding[chp->stack[3]]);
				chp->cont.x=chp->stack[1]+chp->sbx-chp->stack[0];
				chp->cont.y=chp->stack[2]+chp->sby;
				chp->contcharpos=charstart(gstate, StandardEncoding[chp->stack[4]]);
				chp->currstack=chp->stack;	/* clear char operand stack */
				chp->seac=					/* don't return CharWidth any more */
				chp->contchar=TRUE; 		/* continue after this one */
				break;

			case fec_sbw:
#ifdef CHARLOG
	STRFormatPut(line,"SBW %d %d %d %d",chp->stack[0],chp->stack[1],chp->stack[2],chp->stack[3]);
	IOPutString(log,line,NULL);
#endif CHARLOG
				if (!chp->seac)
					endchar=do_width(gstate,(pt **)&comm,chp->stack[2],chp->stack[3]);
				chp->charposx+=chp->sbx=chp->stack[0];
				chp->charposy+=chp->sby=chp->stack[1];
				break;

			case fec_div:
				value=*--chp->currstack;
				*chp->currstack++=*--chp->currstack/value;
				break;

			case fec_callothersubr:
#ifdef TESTFONT
printf("callothersubr %d",currchp->stack[-1]);
#endif
				/* remove othersubr number */
				if (chp->currstack!=chp->stack)
					switch(*--chp->currstack)
					{
					case 3:
						HintClear(chp->hints);
						break;
					case 2:
						chp->currstack=chp->stack;
						break;
					case 1:
						flex=TRUE;
						chp->currstack=chp->stack;
						break;
					case 0:
						do_aline(chp,(pt **)&comm,chp->stack[1],chp->stack[2]);
						flex=FALSE;
						endchar=TRUE;
						break;
					}
				break;

			case fec_pop:
				/* pop value from stack */
				if (chp->currstack!=chp->stack) chp->currstack--;
				break;

			case fec_setcurrentpoint:
				if (chp->currstack!=chp->stack)
				{
#ifdef CHARLOG
	STRFormatPut(line,"POINTCURRSET %d %d",chp->stack[0],chp->stack[1]);
	IOPutString(log,line,NULL);
#endif CHARLOG
					chp->charposx=chp->stack[0];
					chp->charposy=chp->stack[1];
				}
				chp->currstack=chp->stack;		/* clear char operand stack */
				break;
#ifdef TESTFONT
default:
printf("unknown escape command %d",command);
break;
#endif
			}
			break;

		case fc_hsbw:
#ifdef CHARLOG
	STRFormatPut(line,"SSBW %d %d",chp->stack[0],chp->stack[1]);
	IOPutString(log,line,NULL);
#endif CHARLOG
			if (!chp->seac)
				endchar=do_width(gstate,(pt **)&comm,chp->stack[1],0);
			chp->charposx+=chp->sbx=chp->stack[0];
			chp->charposy+=chp->sby=0;
			break;

		case fc_endchar:
			if (chp->contcharpos)
			{
				chp->charpos=chp->contcharpos;
				chp->charposx=chp->charposy=0;
				do_move(chp,(pt **)&comm,chp->cont.x,chp->cont.y);
				chp->contcharpos=FALSE;
			}
			else
			{
#ifdef CHARLOG
	IOPutString(log,"new character",NULL);
	logging=FALSE;
#endif
				endchar=do_endchar(gstate,chp);
				chp->charpos=NULL;
			}
			break;

		case fc_rmoveto:
			if (!flex)
				endchar=do_move(chp,(pt **)&comm,chp->stack[0],chp->stack[1]);
			else
				chp->currstack=chp->stack;		/* clear char operand stack */
			break;

		case fc_hmoveto:
			if (!flex)
				endchar=do_move(chp,(pt **)&comm,chp->stack[0],0);
			else
				chp->currstack=chp->stack;		/* clear char operand stack */
			break;

		case fc_vmoveto:
			if (!flex)
				endchar=do_move(chp,(pt **)&comm,0,chp->stack[0]);
			else
				chp->currstack=chp->stack;		/* clear char operand stack */
			break;

		case fc_rrcurveto:
			endchar=do_curve(chp,(pt **)&comm, chp->stack[0], chp->stack[1], chp->stack[2], chp->stack[3], chp->stack[4], chp->stack[5]);
			break;

		case fc_vhcurveto:
			endchar=do_curve(chp,(pt **)&comm, 0, chp->stack[0], chp->stack[1], chp->stack[2], chp->stack[3], 0);
			break;

		case fc_hvcurveto:
			endchar=do_curve(chp,(pt **)&comm, chp->stack[0], 0, chp->stack[1], chp->stack[2], 0, chp->stack[3]);
			break;
#ifdef TESTFONT
default:
printf("unknown command %d",command);
break;
#endif
		}
	}
	return ERR_OK;
}
